PImage bin;
final static int INTVAL = 100;
final static int FLOATVAL = 101;

public class SecondApplet extends PApplet {
  ControlP5 cp5;

  final static int SINE = 0; //add 0.25 to phase to get cos
  final static int SAWTOOTH = 1;
  final static int SAWTOOTHINVERSE = 2;
  final static int SQUARE = 3; //switches between -1 and 1, that's all
  final static int TRIANGLE = 4;
  final static int TAN = 5;
  final static int TANINVERSE = 6;
  final static int RAMP = 7; //line for half the sequence up to 1, then stay at 1 for the other half
  final static int RAMPINVERSE = 8; //line for half the sequence up to 1, then stay at 1 for the other half
  final static int RAMPUPDOWN = 9; //line up, stay, line down, stay
  final static int RANDOM = 10;
  final static int RANDOMIZE_ALL = 777; //not for wave-generation. meant for low frequency oscillator that calls a function after every beat
  String availableFx[] = {
    "fxASDFPixelSort", 
    "fxDistorter", 
    "fxWahWah", 
    "fxAuEcho", 
    "fxPhaser", 
    "fxWZIP", 
    "fxDarker", 
    "fxBrighter", 
    "fxMove", 
    "fxScale", 
    "fxEpilepsy", 
    "fxPosterize", 
    "fxCopyZoom", 
    "fxDual", 
    "fxScanker", 
    "fxPixelSort", 
    "fxHSB", 
    "fxEcho", 
    "fxSubtleSort", 
    "fxBlackWhite", 
    "fxDrawGenerative", 
    "fxDrawStrokes", 
    "fxFM", 
    "fxGrauzone", 
    "fxSlitScan"
  };
  
  shaderManager shaderManager = new shaderManager();
  class shaderManager {
    void addShader(Shader shader) {
      shaderList.add(shader);
      shader.pos = shaderList.size()-1;
      bricks.add(new Brick(shaderList.size()-1, shaderList.get(shaderList.size()-1).params));
      println("Added " + shader + " to fxList at spot. List length: " + shaderList.size());
    }
    void swapShader(int pos1, int pos2) {
      if (pos1 >= 0 && pos2 < shaderList.size()) {
        Collections.swap(shaderList, pos1, pos2);
        Collections.swap(bricks, pos1, pos2);
        shaderList.get(pos1).pos = pos1;
        shaderList.get(pos2).pos = pos2;
        bricks.get(pos1).setPosition(pos1);
        bricks.get(pos2).setPosition(pos2);
      }
    }
    void removeShader(int pos) {
      if (pos >= 0 && shaderList.size() > pos) { 
        shaderList.remove(pos);
        bricks.remove(pos);
        for (int i = shaderList.size()-1; i > -1; i--) {
          bricks.get(i).setPosition(i);
          shaderList.get(i).pos = i;
        }
      }
    }
    void applyShader() {
      for (int i = 0; i < shaderList.size(); i++) {
        shaderList.get(i).getValuesFromGUI();
        shaderList.get(i).apply();
        if (i > maxFx) break;
      }
    }
  }

  ArrayList<Shader> shaderList = new ArrayList<Shader>();
  ArrayList<Brick> bricks = new ArrayList<Brick>();

  class Brick {
    ArrayList<Slider> slider = new ArrayList<Slider>();
    ArrayList<Param> params = new ArrayList<Param>();
    int position;

    Brick(int position_, ArrayList params_) {
      position = position_;
      params = params_;
      for (int i = 0; i < params.size(); i++) {
        slider.add(new Slider(10, 10+20*i+position*75, 230, 12, params.get(i).minValue, params.get(i).maxValue, params.get(i).name, params.get(i).type)); //add possible waves to slider
      }
    }

    void exchangeSlider(int sliderNo, Param param) {
      slider.get(sliderNo).setMinMax(param.minValue, param.maxValue);
      slider.get(sliderNo).setMode(param.type);
      slider.get(sliderNo).setLabel(param.name);
    }


    void update() {
      for (int i = 0; i < slider.size(); i++) {
        slider.get(i).update();
        params.get(i).value = slider.get(i).value;
      }
    }

    void display() {
      for (int i = 0; i < slider.size(); i++) {
        slider.get(i).display();
      }
    }
  }


  ArrayList<Slider> slider;
  int pressedSlider = -1;

  class Slider {
    int x, y, w, h, mode;
    float minVal, maxVal, value; //should be private but doesnt work lol
    boolean hovering, hovering_anim_btn, animated, randomized;
    String label;
    Oscillator osci;

    Slider(int x_, int y_, int w_, int h_, float minVal_, float maxVal_, String label_, int mode_) {
      x = x_;
      y = y_;
      w = w_;
      h = h_;
      minVal = minVal_;
      maxVal = maxVal_;
      mode = mode_;
      label = label_ + (mode == INTVAL ? " (INT)" : " (FLOAT)");

      osci = new Oscillator(int(random(2, 20)));
      osci.amplitude(minVal, maxVal);

      animated = true;
      randomized = true;

      slider.add(this);
    }

    void setPosX(int x_) {
      x = x_;
    }

    void setMinMax(float min, float max) {
      float cval = map(value, minVal, maxVal, 0, 10000);
      minVal = min;
      maxVal = max;
      osci.amplitude(minVal, maxVal);
      setValue(map(cval, 0, 10000, minVal, maxVal));
    }
    void setLabel(String lab) {
      label = lab + (mode == INTVAL ? " (INT)" : " (FLOAT)");
    }
    void setMode(int mode_) {
      mode = mode_;
    }

    void press(int myId_) {
      pressedSlider = myId_;
      animated = false;
      setValue(map(mouseX, x, x+w, minVal, maxVal));
    }

    void setValue(float value_) {
      value = value_;
      value = constrain(value, minVal, maxVal);
      if (mode == INTVAL) value = int(round(value));
    }

    float getValue() {
      return(value);
    }

    void update() {
      osci.update();

      if (animated) {
        setValue(osci.value);
      }


      if (mouseX >= x && mouseX <= x+w && mouseY >= y && mouseY <= y+h) {
        hovering = true;
      } else {
        hovering = false;
      }

      if (mouseX >= x+w+5 && mouseX <= x+w+5+15 && mouseY >= y && mouseY <= y+h) {
        hovering_anim_btn = true;
      } else {
        hovering_anim_btn = false;
      }
    }

    void display() {
      fill(100+(40*int(hovering)));
      rect(x, y, w, h);
      fill(185+(40*int(hovering)));
      rect(x, y, map(value, minVal, maxVal, 0, w), h);
      fill(255);
      text(label, x+5, y+10, 10);

      if (animated) {
        fill(185+(40*int(hovering_anim_btn)));
      } else {
        fill(100+(40*int(hovering_anim_btn)));
      }
      rect(x+w+5, y, 15, 15);
    }
  }


  public void settings() {
    size(275, 500);
  }

  public void setup() {
    slider = new ArrayList<Slider>(); //brick 

    cp5 = new ControlP5(this);
    List l = Arrays.asList(availableFx);


    cp5.addScrollableList("")
      .setPosition(10, height-110)
      .setSize(80, 100)
      .setBarHeight(0)
      .setItemHeight(20)
      .plugTo(this, "liste")
      .addItems(l)
      .setType(ControlP5.LIST);


    surface.setLocation(285, 240);
    //surface.setSize(1000,100);

    //testslider = new Slider(10, 10, 230, 15, 0, 5, "testslider", INTVAL);

    randomizer = new Oscillator(5); //every 12 seconds;
    randomizer.mode = RANDOMIZE_ALL;
  }



  public void draw() {
    background(68, 68, 68);
    image(bin, width-55, height-85, 40, 40);
    stroke(255);
    strokeWeight(1);
    line(0, height-120, width, height-120);

    randomizer.update();

    //this goes to bricks



    for (int i = 0; i < shaderList.size(); i++) { 
      bricks.get(i).update();
      bricks.get(i).display();

      //for (int j = 0; j < fxList.get(i).shader.params.size(); j++) {
      //Slider s = fxList.get(i).shader.params.get(i);
      //}
    }
  }

  //CP5
  void liste(int n) {
    if (shaderList.size() < maxFx) {
      switch(n) {
        case(0): 
        shaderManager.addShader(new ASDFPIXELSORT()); 
        break;
        case(1): 
        shaderManager.addShader(new DISTORTER()); 
        break;
        /*
        case(2): 
         shaderManager.addShader(new WAHWAH()); 
         break;
         case(3): 
         shaderManager.addShader(new AUECHO()); 
         break;
         case(4): 
         shaderManager.addShader(new PHASER()); 
         break;
         case(0): 
         shaderManager.addShader(new WZIP()); 
         break;
         case(0): 
         shaderManager.addShader(new DARKER()); 
         break;
         case(0): 
         shaderManager.addShader(new BRIGHTER()); 
         break;
         case(0): 
         shaderManager.addShader(new MOVE()); 
         break;
         case(0): 
         shaderManager.addShader(new SCALE()); 
         break;
         case(0): 
         shaderManager.addShader(new EPILEPSY()); 
         break;
         case(0): 
         shaderManager.addShader(new POSTERIZE()); 
         break;
         case(0): 
         shaderManager.addShader(new COPYZOOM()); 
         break;
         case(0): 
         shaderManager.addShader(new DUAL()); 
         break;
         case(0): 
         shaderManager.addShader(new SCANKER()); 
         break;
         case(0): 
         shaderManager.addShader(new PIXELSORT()); 
         break;
         case(0): 
         shaderManager.addShader(new HSB()); 
         break;
         case(0): 
         shaderManager.addShader(new ECHO()); 
         break;
         case(0): 
         shaderManager.addShader(new SUBTLESORT()); 
         break;
         case(0): 
         shaderManager.addShader(new BLACKWHITE()); 
         break;
         case(0): 
         shaderManager.addShader(new DRAWGENERATIVE()); 
         break;
         case(0): 
         shaderManager.addShader(new DRAWSTROKES()); 
         break;
         case(0): 
         shaderManager.addShader(new FM()); 
         break;
         case(0): 
         shaderManager.addShader(new GRAUZONE()); 
         break;
         case(0): 
         shaderManager.addShader(new SLITSCAN()); 
         break;
         */
      default:
        break;
      }
    }
  }




  class Oscillator {
    float speed; //in milliseconds
    int spawnMillis;
    int pauseMillis;
    int beatCount = 0;
    int mode;

    //int function; //function to be called by bang-oscillators
    float position; // between 0 and TWO_PI
    float value; // is returned, constrained between "low" and "high"
    float low = -1.0;
    float high = 1.0;
    boolean playing, pause, valInvert;

    Oscillator(float bpm_) {
      bpm(bpm_);
      play();
    }

    void amplitude(float low_, float high_) {
      low = low_;
      high = high_;
    }
    void frequency(float hertz_) {
      speed = 1000/hertz_;
    }
    void bpm(float bpm_) {
      speed = int(60 / bpm_ * 1000);
    }

    void play() {
      if (pause) {
        float pauseTime = millis() - pauseMillis;
        spawnMillis += pauseTime;
        pause = false;
      } else if (!playing) {
        spawnMillis = millis();
      }
      playing = true;
    }
    void pause() {
      if (playing) {
        pauseMillis = millis();
        pause = true;
        playing = false;
      }
    }
    void stop() {
      beatCount = 0;
      playing = false;
      value = 0;
    }


    void update() {
      if (playing) {
        while (millis() - spawnMillis > speed*beatCount) {
          beatCount++;
          if (mode == RANDOMIZE_ALL) randomize_all();
        }

        float millisToBeat = millis() - spawnMillis - (speed*beatCount);
        position = map(-millisToBeat, speed, 0, 0, TWO_PI);

        switch(mode) {
          case(SINE):
          value = sin(position);
          break;
          case(SAWTOOTH):
          value = ((position % TWO_PI)-PI)/PI;
          break;
          case(SAWTOOTHINVERSE):
          value = (((position % TWO_PI)-PI)/PI)*-1;
          break;
          case(SQUARE):
          value = sin(position > PI ? 1 : -1);
          break;
          case(TRIANGLE):
          value = position-PI < 0 ? ((position % TWO_PI)-PI)/PI : (((position % TWO_PI)-PI)/PI)*-1;
          value+=0.5;
          value*=2;
          if (value == 3.0) value = -1;
          break;
          case(TAN):
          value = tan(position);
          break;
          case(TANINVERSE):
          value = tan(position)*-1;
          break;
          case(RAMP):
          value = sin(position) > 0 ? ((position % TWO_PI)-PI)/PI : 0;
          value+=0.5;
          value*=2;
          if (value == 3.0) value = -1;
          break;
          case(RAMPINVERSE):
          value = sin(position) > 0 ? ((position % TWO_PI)-PI)/PI : 0;
          value+=0.5;
          value*=2;
          if (value == 3.0) value = -1;
          value = value*-1;
          break;
          case(RAMPUPDOWN):
          if (sin(position) <= -1.6666666666) value = -1.6666666666;
          else if (sin(position) < 1.6666666666 && sin(position) > -1.6666666666) value = sin(position);
          else if (sin(position) >= 1.6666666666) value = 1.6666666666;
          value *= 1.5;
          break;
        default:
          break;
        }
        value = constrain(value, -1.0, 1.0);

        value = map(value, -1.0, 1.0, low, high);
        //value = valInvert? map(value, 1.0, -1.0, low, high) : map(value, -1.0, 1.0, low, high);
      }

      //fill(0);
      //ellipse(inc, height/2+value*height/4, 1, 1);
    }
  }


  void randomize_all() {
    for (int i = slider.size()-1; i >= 0; i--) { 
      Slider s = slider.get(i);
      if (random(1) > .5) s.osci.mode = int(round(random(9)));
      if (random(1) > .5) s.osci.play(); 
      else pause();
      if (random(1) > .5) s.osci.bpm(random(2, 10));
    }
  }


  void mousePressed() {
    for (int i = slider.size()-1; i >= 0; i--) { 
      Slider s = slider.get(i);
      if (s.hovering) s.press(i);
      if (s.hovering_anim_btn) s.animated = !s.animated;
    }
  }
  void mouseDragged() {
    if (pressedSlider != -1) {
      Slider s = slider.get(pressedSlider);
      s.setValue(map(mouseX, s.x, s.x+s.w, s.minVal, s.maxVal));
    }
  }

  void mouseReleased() {
    pressedSlider = -1;
  }
}
